/*
** iplog_filter.c - iplog traffic filter routines.
** Copyright (C) 1999 Odin <odin@numb.org>, behe <eric@ojnk.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <netdb.h>
#include <iplog.h>

#include <icmp-compat.h>
#include <iplog_filter.h>

static int get_icmp_code(const u_char *, struct port_range *);
static int get_addr(const u_char *, addr_t *, addr_t *, u_long);
static int get_addrrange(const u_char *, struct addr_entry *, addr_t *, u_long);
static int get_port(const u_char *, int, const u_char *);
static int get_portrange(const u_char *, struct port_range *, const u_char *);
static bool port_match(const struct port_range *, port_t port);
static bool type_match(const struct port_range *, port_t port);
static bool addr_match(const struct addr_entry *, addr_t addr);
static size_t make_pretty(const u_char *buf, u_char *dest, size_t len);

static const u_char *proto_list[] = { "tcp", "udp",	"icmp" };
static u_char reason[64];
static struct filter_data *filters[3];

bool apply_icmp_filter(u_int prot, const struct ip *ip, u_char type)
{
	const struct filter_data *rule, *head = filters[prot];

	for (rule = head ; rule != NULL ; rule = rule->next) {
		if (addr_match(&rule->src, ip->ip_src.s_addr) &&
			addr_match(&rule->dst, ip->ip_dst.s_addr) &&
			type_match(&rule->dport, type))

			return (false ^ rule->not);
	}

	return (false);
}

bool apply_filter(u_int p, const struct ip *ip, port_t sport, port_t dport) {
	const struct filter_data *rule, *head = filters[p];

	for (rule = head ; rule != NULL ; rule = rule->next) {
		if (addr_match(&rule->src, ip->ip_src.s_addr) &&
			addr_match(&rule->dst, ip->ip_dst.s_addr) &&
			port_match(&rule->dport, htons(dport)) &&
			port_match(&rule->sport, htons(sport)))

			return (false ^ rule->not);
	}

	return (false);
}

static bool addr_match(const struct addr_entry *test, addr_t host) {
	if (test->addr == 0 || host == 0 || test->addr == host)
		return (true ^ test->not);

	if (((host ^ test->addr) & test->mask) == 0)
		return (true ^ test->not);

	return (false ^ test->not);
}

static bool type_match(const struct port_range *test, port_t port) {
	if (test->min == TYPE_MATCH_ALL || test->min == port)
		return (true ^ test->not);

	return (false ^ test->not);
}

static bool port_match(const struct port_range *test, port_t port) {
	if (!test->max && !test->min)
		return (true ^ test->not);

	if (test->max == test->min)
		return ((test->max == port) ^ test->not);

	if (!test->max)
		return ((test->min <= port) ^ test->not);
	else if (!test->min)
		return ((test->max >= port) ^ test->not);
	else
		return (((test->max >= port) && (test->min <= port)) ^ test->not);

	return (false ^ test->not);
}

static int get_addr(const u_char *hostname, addr_t *addr,
					addr_t *list, u_long siz)
{
	struct in_addr in;

	if (isdigit(hostname[strlen(hostname) - 1])) {
		u_char *p = (typeof(p)) hostname, i;

		for (i = 0 ; *p != '\0' ; p++)
			if (*p == '.')
				i++;
		if (i != 3) {
			u_char host[MAX_IPLEN];

			xstrncpy(host, hostname, sizeof(host));
			switch (i) {
				case 0:
					xstrncat(host, ".0", sizeof(host));
				case 1:
					xstrncat(host, ".0", sizeof(host));
				case 2:
					xstrncat(host, ".0", sizeof(host));
			}
			inet_aton(host, &in);
			*addr = in.s_addr;
			return (0);
		}
	}

	if (inet_aton(hostname, &in)) {
		*addr = in.s_addr;
		return (0);
	} else {
		struct hostent *host = gethostbyname(hostname);

		if (host != NULL) {
			u_char **p;

			*addr = *((addr_t *) host->h_addr);
			p = (typeof(p)) host->h_addr_list;

			if (list != NULL && *(p + 1) != NULL) {
				u_int i;

				for (i = 0 ; i < siz && *p != NULL ; p++, i++)
					list[i] = *(addr_t *) *p;

				return (i);
			}

			return (0);
		}
	}

	return (-1);
}

static int get_icmp_code(const u_char *ptr, struct port_range *dst) {
	if (*ptr == '!') {
		dst->not = true;
		ptr++;
	}

	if (valid_number(ptr)) {
		int type = strtol(ptr, NULL, 10);

		if (type < 0 || type > 18)
			return (-1);

		dst->min = type;
		return (0);
	}

	if (!strncasecmp(ptr, "ICMP_", 5))
		ptr += 5;

	if (!strncasecmp(ptr, "ECHOREPLY", 5))
		dst->min = ICMP_ECHO_REPLY;
	else if (!strncasecmp(ptr, "ECHO", 2))
		dst->min = ICMP_ECHO;
	else if (!strncasecmp(ptr, "UNREACH", 4))
		dst->min = ICMP_DEST_UNREACHABLE;
	else if (!strncasecmp(ptr, "REDIRECT", 3))
		dst->min = ICMP_REDIRECT;
	else if (!strncasecmp(ptr, "ROUTERADVERT", 9))
		dst->min = ICMP_ROUTER_ADVERT;
	else if (!strncasecmp(ptr, "ROUTERSOLICIT", 9))
		dst->min = ICMP_ROUTER_SOLICIT;
	else if (!strncasecmp(ptr, "TIMXCEED", 5))
		dst->min = ICMP_TIME_EXCEEDED;
	else if (!strncasecmp(ptr, "TIMESTAMP_REPLY", 10))
		dst->min = ICMP_TIMESTAMP_REPLY;
	else if (!strncasecmp(ptr, "TSTAMP", 6))
		dst->min = ICMP_TIMESTAMP;
	else if (!strncasecmp(ptr, "IREQREPLY", 5))
		dst->min =  ICMP_INFO_REPLY;
	else if (!strncasecmp(ptr, "IREQ", 2))
		dst->min = ICMP_INFO_REQUEST;
	else if (!strncasecmp(ptr, "MASKREPLY", 5))
		dst->min = ICMP_ADDRESS_REPLY;
	else if (!strncasecmp(ptr, "MASKREQ", 2))
		dst->min = ICMP_ADDRESS;
	else if (!strncasecmp(ptr, "PARAMPROB", 4))
		dst->min = ICMP_PARAMETER_PROBLEM; 
	else if (!strncasecmp(ptr, "SOURCEQUENCH", 3))
		dst->min = ICMP_SOURCE_QUENCH;
	else {
		snprintf(reason, sizeof(reason), "Unknown icmp type: %s", ptr);
		return (-1);
	}

	return (0);
}

static int get_port(const u_char *s, int wildcard, const u_char *proto) {
	int temp;

	if (!strcasecmp(s, EVERYTHING))
		temp = wildcard;
	else {
		if (valid_number(s))
			temp = strtol(s, NULL, 10);
		else {
			struct servent *se = getservbyname(s, proto);
			if (se == NULL)
				return (-1);

			temp = htons(se->s_port);
		}
	}

	return (temp);
}

static int get_addrrange(const u_char *b, struct addr_entry *addr,
						addr_t *list, u_long siz)
{
	u_char *p, *q;
	int i = 0;

	if (*b == '!') {
		addr->not = true;
		++b;
	}
	
	p = strchr(b, '/');
	if (p != NULL) {
		*p++ = '\0';
		for (q = p ; *q != '\0' ; q++) {
			if (!isdigit(*q)) {
				i = 1;
				break;
			}
		}
			
		if (i == 1) {
			if (get_addr(p, &addr->mask, NULL, 0) == -1) {
				snprintf(reason, sizeof(reason), "Bad mask: %s", b);
				return (-1);
			}
		} else {
			if (!strcasecmp(p, EVERYTHING))
				addr->mask = 1;
			else if (valid_number(p)) {
				i = strtoul(p, NULL, 10);
				addr->mask = (1 << i) - 1;
			} else {
				snprintf(reason, sizeof(reason), "Bad mask: %s", b);
				return (-1);
			}
		}
	} else
		addr->mask = MASK_MATCH_ONE;

	if (!strcasecmp(b, EVERYTHING))
		addr->addr = 0;
	else {
		i = get_addr(b, &addr->addr, list, siz);
		if (i == -1) {
			snprintf(reason, sizeof(reason), "No such host: %s", b);
			return (-1);
		}
		return (i);
	}
	
	return (0);
}

static int get_portrange(const u_char *b, struct port_range *port, const u_char *proto) {
	u_char *p;

	if (*b == '!') {
		port->not = true;
		++b;
	} else
		port->not = false;

	p = strchr(b, ':');
	if (p != NULL)
		*p++ = '\0';

	port->min = get_port(b, MIN_PORT, proto);

	if (p != NULL) {
		/* Allow for stuff like 6000: */
		if (*p == '\0')
			port->max = MAX_PORT;
		else
			port->max = get_port(p, MAX_PORT, proto);
	} else 
		port->max = port->min;

	if ((port->min & MAX_PORT) != port->min ||
		(port->max & MAX_PORT) != port->max)
	{
		snprintf(reason, sizeof(reason),
			"Error: Ports must be between 1 and 65535");
		return (-1);
	}

	return (0);
}

void read_filters(const u_char *filename) {
	FILE *fp;
	bool tty = isatty(STDERR_FILENO);
	u_int i, z;
	struct filter_data data;
	u_char *p = NULL, *q = NULL, *b = NULL, line = 0, let = 0;
	u_char buf[1024], fbuf[1024], proto = 0;
	void *(*atl)(const void *, void **, size_t) = NULL;
	addr_t dadr[32], dadr_c = 0, sadr[32], sadr_c = 0;

	fp = fopen(filename, "r");

	if (fp == NULL)
		return;

	goto begin;

ojnk:
	if (let) {
		if (reason[0] == '\0')
			xstrncpy(reason, "format error", sizeof(reason));

		if (tty == true)
			fprintf(stderr, "%s:%d: %s\n", filename, line, reason);
		else
			mysyslog(PRIORITY, "%s:%d: %s", filename, line, reason);
	}
	reason[0] = '\0';

begin:
	while (get_line(fp, buf, sizeof(buf) - 1) != EOF) {
		++line;

		let = make_pretty(buf, fbuf, sizeof(fbuf));
		if (let == 0)
			continue;

		data.sport.min = MIN_PORT;
		data.sport.max = MAX_PORT;
		data.sport.not = false;
		data.dport.min = MIN_PORT;
		data.dport.max = MAX_PORT;
		data.dport.not = false;
		data.src.addr = ADDR_MATCH_ALL;
		data.src.mask = MASK_MATCH_ONE;
		data.src.not = false;
		data.dst.addr = ADDR_MATCH_ALL;
		data.dst.mask = MASK_MATCH_ONE;
		data.dst.not = false;
		data.not = false;

		for (p = fbuf, z = 0, i = 0 ; p != NULL ;) {
			q = p;
			p = strchr(q, ' ');
			if (p != NULL)
				*p++ = '\0';

			if (i == 0) {
				if (!strcasecmp(q, "log")) {
					atl = list_prepend;
					data.not = false;
				} else if (!strcasecmp(q, "ignore")) {
					atl = list_append;
					data.not = true;
				} else {
					snprintf(reason, sizeof(reason),
						"First field must be either \"log\" or \"ignore\". "
						"  You put \"%s\".", q);
					goto ojnk;
				}
				++i;
				continue;
			}
			
			if (i == 1) {
				if (!strcasecmp(q, "tcp"))
					proto = FIL_TCP;
				else if (!strcasecmp(q, "udp"))
					proto = FIL_UDP;
				else if (!strcasecmp(q, "icmp")) {
					data.dport.min = TYPE_MATCH_ALL;
					proto = FIL_ICMP;
				} else {
					snprintf(reason, sizeof(reason),
						"Bad protocol: %s", q);
					goto ojnk;
				}
				++i;
				continue;
			}

			b = q;
			q = p;
			p = strchr(q, ' ');
			if (p != NULL)
				*p++ = '\0';

			if (proto != FIL_ICMP) {
				if (!strcasecmp(b, "dport"))
					get_portrange(q, &data.dport, proto_list[proto]);
				else if (!strcasecmp(b, "sport"))
					get_portrange(q, &data.sport, proto_list[proto]);
			} else if (!strcasecmp(b, "type")) {
				if (get_icmp_code(q, &data.dport) == -1)
					goto ojnk;
			}

			if (!strcasecmp(b, "from"))
				sadr_c = get_addrrange(q, &data.src, sadr, sizeof(sadr));
			else if (!strcasecmp(b, "to"))
				dadr_c = get_addrrange(q, &data.dst, dadr, sizeof(dadr));
			else if (strcasecmp(b, "dport") && strcasecmp(b, "sport") &&
					strcasecmp(b, "type"))
			{
				snprintf(reason, sizeof(reason), "Unknown keyword: \"%s\"", b);
				goto ojnk;
			}
		}

		if (sadr_c != 0 || dadr_c != 0) {
			u_int b, s, i, j;
			addr_t *bp, *sp, *bl, *sl;

			if (sadr_c > dadr_c) {
				b = sadr_c;
				s = dadr_c;
				bp = &data.src.addr;
				sp = &data.dst.addr;
				bl = sadr;
				sl = dadr;
			} else {
				b = dadr_c;
				s = sadr_c;
				bp = &data.dst.addr;
				sp = &data.src.addr;
				bl = dadr;
				sl = sadr;
			}

			for (i = 0 ; i < b ; i++) {
				*bp = bl[i];
				for (j = 0 ; j < s ; j++) {
					*sp = sl[j];
					atl(&data, (void **) &filters[proto], sizeof(data));
				}
				if (s == 0)
					atl(&data, (void **) &filters[proto], sizeof(data));
			}

		} else
			atl(&data, (void **) &filters[proto], sizeof(data));
	}

	fclose(fp);
}

static size_t make_pretty(const u_char *buf, u_char *dest, size_t len) {
	size_t i = 0;
	u_char c, l = 0;

	while (isspace(*buf))
		++buf;

	while (i < len && *buf != '\0' && *buf != '#' && *buf != '\n') {
		c = *buf++;
		switch (c) {
			case '\t':
				c = ' ';
				/* FALLTHROUGH */
			case ' ':
				if (i == 0 || l == ' ')
					break;
				/* FALLTHROUGH */
			case '!':
				dest[i++] = c;
				l = ' ';
				break;	
			default:
				dest[i++] = l = c;
		}
	}

	if (i > 0 && dest[i - 1] == ' ')
		dest[i - 1] = '\0';
	else
		dest[i] = '\0';

	return (i);
}

void add_dns_ignore_rules(void) {
	struct filter_data temp;
	u_char buf[1024], nbuf[1024], *p;
	FILE *fp;
	addr_t list[32], a_c;

	fp = fopen(NS_FILE, "r");
	if (fp == NULL)
		return;

	temp.sport.min = 53;
	temp.sport.max = 53;
	temp.dport.min = MIN_PORT;
	temp.dport.max = MAX_PORT;
	temp.src.not = false;
	temp.src.mask = MASK_MATCH_ONE;
	temp.dst.addr = ADDR_MATCH_ALL;
	temp.dst.mask = MASK_MATCH_ONE;
	temp.dst.not = false;
	temp.not = true;
	
	while (get_line(fp, buf, sizeof(buf) - 1) != EOF) {
		if (make_pretty(buf, nbuf, sizeof(nbuf)) == 0)
			continue;

		p = strchr(nbuf, ' ');
		if (p == NULL)
			continue;
		*p++ = '\0';

		if (strcmp(nbuf, "nameserver"))
			continue;

		a_c = get_addr(p, &temp.src.addr, list, sizeof(list));
		list_prepend(&temp, (void **) &filters[FIL_UDP], sizeof(temp));
		if (a_c != 0) {
			u_int i;

			for (i = 0 ; i < a_c ; i++) {
				temp.src.addr = list[i];
				list_prepend(&temp, (void **) &filters[FIL_UDP], sizeof(temp));
			}
		}
	}

	fclose(fp);
}

#ifdef HAVE_PTHREAD_CANCEL
void destroy_filter_list(u_int prot) {
	list_destroy(filters[prot], NULL);
	filters[prot] = NULL;
}
#endif

/* vim:ts=4:sw=8:tw=0 */
